package ru.dbotthepony.kstarbound.defs.item

import com.google.gson.Gson
import com.google.gson.JsonArray
import com.google.gson.JsonElement
import com.google.gson.JsonNull
import com.google.gson.JsonObject
import com.google.gson.JsonPrimitive
import com.google.gson.JsonSyntaxException
import com.google.gson.TypeAdapter
import com.google.gson.annotations.JsonAdapter
import com.google.gson.stream.JsonReader
import com.google.gson.stream.JsonWriter
import org.classdump.luna.ByteString
import org.classdump.luna.LuaRuntimeException
import org.classdump.luna.Table
import org.classdump.luna.TableFactory
import ru.dbotthepony.kommons.gson.consumeNull
import ru.dbotthepony.kommons.gson.contains
import ru.dbotthepony.kommons.util.KOptional
import ru.dbotthepony.kstarbound.lua.StateMachine
import ru.dbotthepony.kstarbound.lua.from
import ru.dbotthepony.kstarbound.lua.toJsonObject
import ru.dbotthepony.kommons.gson.get
import ru.dbotthepony.kommons.gson.value
import ru.dbotthepony.kommons.io.StreamCodec
import ru.dbotthepony.kommons.io.readVarLong
import ru.dbotthepony.kommons.io.writeBinaryString
import ru.dbotthepony.kommons.io.writeVarLong
import ru.dbotthepony.kstarbound.Registries
import ru.dbotthepony.kstarbound.Registry
import ru.dbotthepony.kstarbound.Starbound
import ru.dbotthepony.kstarbound.VersionRegistry
import ru.dbotthepony.kstarbound.defs.item.api.IItemDefinition
import ru.dbotthepony.kstarbound.io.readInternedString
import ru.dbotthepony.kstarbound.item.ItemStack
import ru.dbotthepony.kstarbound.json.readJsonElement
import ru.dbotthepony.kstarbound.json.writeJsonElement
import java.io.DataInputStream
import java.io.DataOutputStream
import java.util.function.Supplier

private val EMPTY_JSON = JsonObject()

fun ItemDescriptor(data: JsonElement): ItemDescriptor {
	if (data is JsonPrimitive) {
		return ItemDescriptor(data.asString, 1L)
	} else if (data is JsonArray) {
		val name = data[0].asString
		val count = data.get(1, 1L)
		val parameters = data.get(2, ::JsonObject)
		return ItemDescriptor(name, count, parameters)
	} else if (data is JsonObject) {
		if ("id" in data && "version" in data && "content" in data) {
			// loading versioned json from original engine
			if (data["id"].asString != "Item")
				throw JsonSyntaxException("Expected id to be 'Item', ${data["id"]} given")

			return ItemDescriptor(data["content"])
		} else {
			// loading regular json
			val name = (data.get("name") ?: data.get("item") ?: throw JsonSyntaxException("Missing item name")).asString
			val count = data.get("count", 1L)
			val parameters = data.get("parameters") { data.get("parameters", ::JsonObject) }
			return ItemDescriptor(name, count, parameters)
		}
	} else if (data is JsonNull) {
		return ItemDescriptor.EMPTY
	} else {
		throw JsonSyntaxException("Invalid item descriptor: $data")
	}
}

fun ItemDescriptor(data: Table, stateMachine: StateMachine): Supplier<ItemDescriptor> {
	val name = stateMachine.index(data, 1L, "name", "item")
	val count = stateMachine.optionalIndex(data, 2L, "count")
	val parameters = stateMachine.optionalIndex(data, 3L, "parameters", "data")
	var result: KOptional<ItemDescriptor> = KOptional.empty()

	stateMachine.add {
		val iname = name.get()
		val icount = count.get().orElse(1L)
		val iparameters = parameters.get().map { if (it is Table) it.toJsonObject() else throw LuaRuntimeException("Invalid item descriptor parameters ($it)") }.orElse { JsonObject() }

		if (iname !is ByteString) throw LuaRuntimeException("Invalid item descriptor name (${iname})")
		if (icount !is Number) throw LuaRuntimeException("Invalid item descriptor count (${icount})")

		result = KOptional(ItemDescriptor(iname.decode(), icount.toLong(), iparameters))
	}

	return Supplier { result.value }
}

fun ItemDescriptor(stream: DataInputStream): ItemDescriptor {
	val name = stream.readInternedString()
	val count = stream.readVarLong()
	val params = stream.readJsonElement()

	if (name == "")
		return ItemDescriptor.EMPTY

	return ItemDescriptor(name, count, if (params === JsonNull.INSTANCE || params == EMPTY_JSON) EMPTY_JSON else params as JsonObject)
}

/**
 * Describes a prototype of item
 *
 * [parameters] is considered to be immutable and should not be modified
 * directly (must be copied for mutable context)
 */
@JsonAdapter(ItemDescriptor.Adapter::class)
data class ItemDescriptor(
	val name: String,
	val count: Long,
	val parameters: JsonObject = EMPTY_JSON
) {
	constructor(ref: Registry.Ref<IItemDefinition>, count: Long, parameters: JsonObject) : this(ref.key.left(), count, parameters)

	val isEmpty get() = count <= 0L || name == "" || ref.isEmpty
	val ref by lazy { if (name == "") Registries.items.emptyRef else Registries.items.ref(name) }

	override fun toString(): String {
		return "ItemDescriptor[$name, $count, $parameters]"
	}

	fun makeStack(): ItemStack {
		return ItemStack.create(this)
	}

	private fun toJsonStruct() = JsonObject().also {
		it.add("name", JsonPrimitive(name))
		it.add("count", JsonPrimitive(count))
		it.add("parameters", parameters.deepCopy())
	}

	fun toJson(): JsonObject? {
		if (Starbound.IS_STORE_JSON) {
			return VersionRegistry.make("Item", toJsonStruct()).toJson()
		} else {
			if (isEmpty) {
				return null
			}

			return toJsonStruct()
		}
	}

	fun toTable(allocator: TableFactory): Table? {
		if (isEmpty) {
			return null
		}

		return allocator.newTable(0, 3).also {
			it.rawset("name", name)
			it.rawset("count", count)
			it.rawset("parameters", allocator.from(parameters))
		}
	}

	fun write(stream: DataOutputStream) {
		stream.writeBinaryString(name)
		stream.writeVarLong(count.coerceAtLeast(0L))
		stream.writeJsonElement(parameters)
	}

	class Adapter(gson: Gson) : TypeAdapter<ItemDescriptor>() {
		private val elements = gson.getAdapter(JsonElement::class.java)

		override fun write(out: JsonWriter, value: ItemDescriptor) {
			if (value.isEmpty)
				out.nullValue()
			else
				out.value(value.toJson())
		}

		override fun read(`in`: JsonReader): ItemDescriptor {
			return ItemDescriptor(elements.read(`in`))
		}
	}

	companion object {
		val EMPTY = ItemDescriptor("", 0)
		val CODEC = StreamCodec.Impl(::ItemDescriptor, { a, b -> b.write(a) })
	}
}
