package ru.dbotthepony.kstarbound.network.syncher

import ru.dbotthepony.kommons.io.StreamCodec
import ru.dbotthepony.kommons.util.Listenable
import ru.dbotthepony.kommons.util.ListenableDelegate
import java.io.DataInputStream
import java.io.DataOutputStream
import java.util.*
import java.util.function.Consumer

open class BasicNetworkedElement<TYPE, LEGACY>(private var value: TYPE, protected val codec: StreamCodec<TYPE>, protected val legacyCodec: StreamCodec<LEGACY>, protected val toLegacy: (TYPE) -> LEGACY, protected val fromLegacy: (LEGACY) -> TYPE) : NetworkedElement(), ListenableDelegate<TYPE> {
	protected val valueListeners = Listenable.Impl<TYPE>()
	protected val queue = LinkedList<Pair<Double, TYPE>>()
	protected var isInterpolating = false
	protected var currentTime = 0.0

	override fun toString(): String {
		return "BasicNetworkedElement[$value, isInterpolating=$isInterpolating, currentTime=$currentTime]"
	}

	override fun accept(t: TYPE) {
		if (t != value) {
			value = t
			queue.clear()
			bumpVersion()
			valueListeners.accept(t)
		}
	}

	override fun addListener(listener: Consumer<TYPE>): Listenable.L {
		return valueListeners.addListener(listener)
	}

	override fun get(): TYPE {
		return value
	}

	override fun readInitial(data: DataInputStream, isLegacy: Boolean) {
		val old = value
		value = if (isLegacy) fromLegacy(legacyCodec.read(data)) else codec.read(data)
		queue.clear()
		bumpVersion()

		if (value != old) {
			valueListeners.accept(value)
		}
	}

	override fun writeInitial(data: DataOutputStream, isLegacy: Boolean) {
		if (isLegacy) {
			if (queue.isNotEmpty()) {
				legacyCodec.write(data, toLegacy(queue.last.second))
			} else {
				legacyCodec.write(data, toLegacy(value))
			}
		} else {
			if (queue.isNotEmpty()) {
				codec.write(data, queue.last.second)
			} else {
				codec.write(data, value)
			}
		}
	}

	override fun readDelta(data: DataInputStream, interpolationDelay: Double, isLegacy: Boolean) {
		val read = if (isLegacy) fromLegacy(legacyCodec.read(data)) else codec.read(data)
		bumpVersion()

		if (isInterpolating) {
			// Only append an incoming delta to our pending value list if the incoming
			// step is forward in time of every other pending value.  In any other
			// case, this is an error or the step tracking is wildly off, so just clear
			// any other incoming values.
			val actualDelay = interpolationDelay + currentTime

			if (interpolationDelay > 0.0 && (queue.isEmpty() || queue.last.first <= actualDelay)) {
				queue.add(actualDelay to read)
			} else {
				value = read
				queue.clear()
				valueListeners.accept(read)
			}
		} else {
			value = read
			valueListeners.accept(read)
		}
	}

	override fun writeDelta(data: DataOutputStream, remoteVersion: Long, isLegacy: Boolean) {
		writeInitial(data, isLegacy)
	}

	override fun readBlankDelta(interpolationDelay: Double) {
		// TODO: original engine doesn't override this, is this intentional?
		// tickInterpolation(interpolationDelay)
	}

	override fun enableInterpolation(extrapolation: Double) {
		if (!isInterpolating) {
			isInterpolating = true
			queue.clear()
		}
	}

	override fun disableInterpolation() {
		if (isInterpolating) {
			isInterpolating = false

			if (queue.isNotEmpty()) {
				value = queue.last.second
				valueListeners.accept(value)
			}

			queue.clear()
		}
	}

	override fun tickInterpolation(delta: Double) {
		require(delta >= 0.0) { "Negative interpolation delta: $delta" }
		currentTime += delta

		while (queue.isNotEmpty() && queue.first.first <= currentTime) {
			value = queue.removeFirst().second
			valueListeners.accept(value)
		}
	}
}
