package ru.dbotthepony.kstarbound.server.world

import it.unimi.dsi.fastutil.objects.ObjectArraySet
import ru.dbotthepony.kommons.arrays.Object2DArray
import ru.dbotthepony.kommons.vector.Vector2d
import ru.dbotthepony.kommons.vector.Vector2i
import ru.dbotthepony.kstarbound.defs.tile.BuiltinMetaMaterials
import ru.dbotthepony.kstarbound.defs.tile.TileDamage
import ru.dbotthepony.kstarbound.defs.tile.TileDamageResult
import ru.dbotthepony.kstarbound.defs.tile.TileDamageType
import ru.dbotthepony.kstarbound.defs.tile.isEmptyTile
import ru.dbotthepony.kstarbound.defs.tile.isNullTile
import ru.dbotthepony.kstarbound.defs.tile.supportsModifier
import ru.dbotthepony.kstarbound.network.LegacyNetworkCellState
import ru.dbotthepony.kstarbound.network.packets.clientbound.TileDamageUpdatePacket
import ru.dbotthepony.kstarbound.world.CHUNK_SIZE
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.TileHealth
import ru.dbotthepony.kstarbound.world.api.AbstractCell
import ru.dbotthepony.kstarbound.world.api.ImmutableCell
import ru.dbotthepony.kstarbound.world.api.TileColor
import ru.dbotthepony.kstarbound.world.entities.AbstractEntity

class ServerChunk(world: ServerWorld, pos: ChunkPos) : Chunk<ServerWorld, ServerChunk>(world, pos) {
	/**
	 * Determines the state chunk is in, chunks are written to persistent storage
	 * only if they are in [FULL] state to avoid partially loaded or partially generated
	 * chunks from making its way into persistent storage.
	 */
	enum class State {
		FRESH, // Nothing is loaded

		TILES,
		MICRO_DUNGEONS,
		CAVE_LIQUID,
		TILE_ENTITIES,
		ENTITIES,

		FULL; // indicates everything has been loaded
	}

	var state: State = State.FRESH
		private set

	fun bumpState(newState: State) {
		require(newState >= state) { "Tried to downgrade $this state from $state to $newState" }

		if (newState >= State.ENTITIES) {
			this.state = State.FULL
		} else {
			this.state = newState
		}
	}

	fun copyCells(): Object2DArray<ImmutableCell> {
		if (cells.isInitialized()) {
			return Object2DArray(cells.value)
		} else {
			return Object2DArray(CHUNK_SIZE, CHUNK_SIZE, AbstractCell.EMPTY)
		}
	}

	data class DamageResult(val result: TileDamageResult, val health: TileHealth? = null, val stateBefore: AbstractCell? = null)

	fun damageTile(pos: Vector2i, isBackground: Boolean, sourcePosition: Vector2d, damage: TileDamage, source: AbstractEntity? = null): DamageResult {
		if (!cells.isInitialized()) {
			return DamageResult(TileDamageResult.NONE)
		}

		val cell = cells.value[pos.x, pos.y]

		if (cell.isIndestructible || cell.tile(isBackground).material.value.isMeta) {
			return DamageResult(TileDamageResult.NONE)
		}

		var damage = damage
		var result = TileDamageResult.NORMAL

		if (cell.dungeonId in world.protectedDungeonIDs) {
			damage = damage.copy(type = TileDamageType.PROTECTED)
			result = TileDamageResult.PROTECTED
		}

		val health = (if (isBackground) tileHealthBackground else tileHealthForeground).value[pos.x, pos.y]
		val tile = cell.tile(isBackground)

		val params = if (!damage.type.isPenetrating && tile.modifier != null && tile.modifier!!.value.breaksWithTile) {
			tile.material.value.actualDamageTable + tile.modifier!!.value.actualDamageTable
		} else {
			tile.material.value.actualDamageTable
		}

		health.damage(params, sourcePosition, damage)
		subscribers.forEach { it.onTileHealthUpdate(pos.x, pos.y, isBackground, health) }

		if (health.isDead) {
			if (isBackground) {
				damagedTilesBackground.remove(pos)
			} else {
				damagedTilesForeground.remove(pos)
			}

			val copyHealth = health.copy()
			val mCell = cell.mutable()
			val mTile = mCell.tile(isBackground)

			mTile.material = BuiltinMetaMaterials.EMPTY
			mTile.color = TileColor.DEFAULT
			mTile.hueShift = 0f

			if (tile.modifier.value.breaksWithTile) {
				mTile.modifier = BuiltinMetaMaterials.EMPTY_MOD
			}

			setCell(pos.x, pos.y, mCell.immutable())
			health.reset()
			return DamageResult(result, copyHealth, cell)
		} else {
			if (isBackground) {
				damagedTilesBackground.add(pos)
			} else {
				damagedTilesForeground.add(pos)
			}

			return DamageResult(result, health, cell)
		}
	}

	private val damagedTilesForeground = ObjectArraySet<Vector2i>()
	private val damagedTilesBackground = ObjectArraySet<Vector2i>()

	fun tileDamagePackets(): List<TileDamageUpdatePacket> {
		val result = ArrayList<TileDamageUpdatePacket>()

		if (tileHealthBackground.isInitialized()) {
			val tileHealthBackground = tileHealthBackground.value

			for (x in 0 until CHUNK_SIZE) {
				for (y in 0 until CHUNK_SIZE) {
					val health = tileHealthBackground[x, y]

					if (!health.isHealthy) {
						result.add(TileDamageUpdatePacket(pos.tileX + x, pos.tileY + y, true, health))
					}
				}
			}
		}

		if (tileHealthForeground.isInitialized()) {
			val tileHealthForeground = tileHealthForeground.value

			for (x in 0 until CHUNK_SIZE) {
				for (y in 0 until CHUNK_SIZE) {
					val health = tileHealthForeground[x, y]

					if (!health.isHealthy) {
						result.add(TileDamageUpdatePacket(pos.tileX + x, pos.tileY + y, false, health))
					}
				}
			}
		}

		return result
	}

	override fun tick() {
		if (state != State.FULL)
			return

		super.tick()

		if (cells.isInitialized() && (damagedTilesBackground.isNotEmpty() || damagedTilesForeground.isNotEmpty())) {
			val tileHealthBackground = tileHealthBackground.value
			val tileHealthForeground = tileHealthForeground.value
			val cells = cells.value

			damagedTilesBackground.removeIf { (x, y) ->
				val health = tileHealthBackground[x, y]
				val result = !health.tick(cells[x, y].background.material.value.actualDamageTable)
				subscribers.forEach { it.onTileHealthUpdate(x, y, true, health) }
				result
			}

			damagedTilesForeground.removeIf { (x, y) ->
				val health = tileHealthForeground[x, y]
				val result = !health.tick(cells[x, y].foreground.material.value.actualDamageTable)
				subscribers.forEach { it.onTileHealthUpdate(x, y, false, health) }
				result
			}
		}
	}

	fun legacyNetworkCells(): Object2DArray<LegacyNetworkCellState> {
		val width = (world.geometry.size.x - pos.tileX).coerceAtMost(CHUNK_SIZE)
		val height = (world.geometry.size.y - pos.tileY).coerceAtMost(CHUNK_SIZE)

		if (cells.isInitialized()) {
			val cells = cells.value
			return Object2DArray(width, height) { a, b -> cells[a, b].toLegacyNet() }
		} else {
			return Object2DArray(width, height, LegacyNetworkCellState.NULL)
		}
	}

	fun prepareCells() {
		val cells = cells.value

		val width = (world.geometry.size.x - pos.tileX).coerceAtMost(cells.columns)
		val height = (world.geometry.size.y - pos.tileY).coerceAtMost(cells.rows)

		for (x in 0 until width) {
			for (y in 0 until height) {
				val info = world.template.cellInfo(pos.tileX + x, pos.tileY + y)

				val state = cells[x, y].mutable()

				state.blockBiome = info.blockBiomeIndex
				state.envBiome = info.environmentBiomeIndex

				if (state.foreground.material.isNullTile) {
					state.foreground.material = info.foreground.entry ?: BuiltinMetaMaterials.EMPTY
					state.foreground.color = TileColor.DEFAULT
					state.foreground.hueShift = info.blockBiome?.hueShift(state.foreground.material) ?: 0f

					if (state.foreground.material.supportsModifier(info.foregroundMod)) {
						state.foreground.modifier = info.foregroundMod.entry ?: BuiltinMetaMaterials.EMPTY_MOD
						state.foreground.hueShift = info.blockBiome?.hueShift(info.foregroundMod) ?: 0f
					}
				}

				if (state.background.material.isNullTile) {
					state.background.material = info.background.entry ?: BuiltinMetaMaterials.EMPTY
					state.background.color = TileColor.DEFAULT
					state.background.hueShift = info.blockBiome?.hueShift(state.background.material) ?: 0f

					if (state.background.material.supportsModifier(info.backgroundMod)) {
						state.background.modifier = info.backgroundMod.entry ?: BuiltinMetaMaterials.EMPTY_MOD
						state.background.hueShift = info.blockBiome?.hueShift(info.backgroundMod) ?: 0f
					}
				}

				if (!state.foreground.material.isEmptyTile) {
					state.liquid.reset()
				} else if (y < info.oceanLiquidLevel && info.oceanLiquid.isPresent && !info.oceanLiquid.value!!.isMeta) {
					val pressure = (info.oceanLiquidLevel - y).toFloat()

					if (state.background.material.isEmptyTile) {
						state.liquid.pressure = pressure
						state.liquid.isInfinite = true
						state.liquid.state = info.oceanLiquid.entry!!
					} else if (info.encloseLiquids) {
						state.liquid.pressure = pressure
						state.liquid.isInfinite = false
						state.liquid.level = 1f
						state.liquid.state = info.oceanLiquid.entry!!
					}
				}

				cells[x, y] = state.immutable()
			}
		}
	}
}
